import inspect
import json
import logging
import re
from functools import wraps
from os import path
from typing import Callable, Dict, Generator, List, Optional, Sequence, Tuple, Type

from flask import Blueprint, Flask, Response, jsonify, render_template, request, send_file
from werkzeug.exceptions import HTTPException

from apistrap.errors import ApiClientError, ApiServerError, UnsupportedMediaTypeError
from apistrap.extension import Apistrap, ErrorHandler, SecurityScheme
from apistrap.operation_wrapper import OperationWrapper
from apistrap.schemas import ErrorResponse
from apistrap.types import FileResponse
from apistrap.utils import format_exception, resolve_fw_decl

SecurityEnforcer = Callable[[Sequence[str]], None]


class FlaskOperationWrapper(OperationWrapper):
    URL_FILTER_MAP = {"string": str, "int": int, "float": float, "path": str}

    def __init__(
        self, extension: Apistrap, function: Callable, decorators: Sequence[object], url_rule: str, method: str
    ):
        self.url_rule = url_rule
        self.method = method
        super().__init__(extension, function, decorators)

    def _enforce_security(self):
        error = None

        for security_scheme, required_scopes in self._get_required_scopes():
            try:
                # If any enforcer passes without throwing, the user is authenticated
                self._extension.security_enforcers[security_scheme](required_scopes)
                return
            except Exception as e:
                error = e
        else:
            if error is not None:
                raise error

    def get_decorated_function(self):
        @wraps(self._wrapped_function)
        def wrapper(*args, **kwargs):
            self._enforce_security()

            if self.accepts_body:
                data = self._load_request_body_primitive()
                kwargs.update(self._load_request_body(data))

            # path parameters are already handled by Flask (and they should be in args/kwargs)

            for name, param_type in self._query_parameters.items():
                if name in request.args.keys():
                    kwargs[name] = param_type(request.args[name])
                elif self._signature.parameters[name].default == inspect.Parameter.empty:
                    raise ApiClientError(f"Missing query parameter `{name}`")

            response, code, mimetype = self._postprocess_response(self._wrapped_function(*args, **kwargs))

            if self.is_raw_response(response):
                return response, code
            if isinstance(response, FileResponse):
                return send_file(
                    filename_or_fp=response.filename_or_fp,
                    mimetype=mimetype or response.mimetype,
                    as_attachment=response.as_attachment,
                    attachment_filename=response.attachment_filename,
                    add_etags=response.add_etags,
                    cache_timeout=response.cache_timeout,
                    conditional=response.conditional,
                    last_modified=response.last_modified,
                )

            response = jsonify(response.to_primitive())
            response.status_code = code
            return response

        return wrapper

    def _load_request_body_primitive(self) -> dict:
        if request.content_type == "application/json":
            try:
                if request.json is None or isinstance(request.json, str):
                    raise ApiClientError("The request body must be a JSON object")
            except json.decoder.JSONDecodeError as ex:
                raise ApiClientError("The request body must be a JSON object") from ex

            return request.json
        elif request.content_type in ("application/x-www-form-urlencoded", "multipart/form-data"):
            return request.form

        raise UnsupportedMediaTypeError()

    def _get_path_parameters(self) -> Generator[Tuple[str, Type], None, None]:
        for param in re.findall("(<([^<>]*:)?([^<>]*)>)", self.url_rule):
            url_filter = param[1].rstrip(":") if param[1] is not None else None
            name = param[2]

            param_type = str

            if url_filter is not None:
                param_type = self.URL_FILTER_MAP.get(url_filter, param_type)

            param_refl: inspect.Parameter = self._signature.parameters[name]

            if param_refl.annotation != inspect.Parameter.empty and param_refl.annotation is not None:
                param_type = resolve_fw_decl(self._wrapped_function, param_refl.annotation)

            yield name, param_type

    def is_raw_response(self, response: object) -> bool:
        return isinstance(response, Response)


class FlaskApistrap(Apistrap):
    def __init__(self):
        super().__init__()
        self._app: Flask = None
        self._specs_extracted = False
        self._operations: Optional[List[FlaskOperationWrapper]] = None
        self.security_enforcers: Dict[SecurityScheme, SecurityEnforcer] = {}

        self._default_error_handlers = (
            ErrorHandler(HTTPException, lambda exc_type: exc_type.code, self.http_error_handler),
            ErrorHandler(UnsupportedMediaTypeError, 415, self.error_handler),
            ErrorHandler(ApiClientError, 400, self.error_handler),
            ErrorHandler(ApiServerError, 500, self.internal_error_handler),
            ErrorHandler(Exception, 500, self.internal_error_handler),
        )

    def init_app(self, app: Flask):
        """
        Bind the extension to a Flask instance.

        :param app: the Flask instance
        """
        self._app = app
        blueprint = Blueprint("apistrap", __name__, template_folder=path.join(path.dirname(__file__), "templates"))

        if self.spec_url is not None:
            blueprint.route(self.spec_url, methods=["GET"])(self._get_spec)

        if self.spec_url is not None and self.ui_url is not None:
            blueprint.route(self.ui_url)(self._get_ui)
            blueprint.route(self.ui_url + "/")(self._get_ui)

        if self.spec_url is not None and self.redoc_url is not None:
            blueprint.route(self.redoc_url)(self._get_redoc)
            blueprint.route(self.redoc_url + "/")(self._get_redoc)

        app.register_blueprint(blueprint)
        if self.has_error_handlers:
            app.register_error_handler(Exception, self._error_handler)

        app.before_first_request_funcs.append(self._decorate_view_handlers)

    def add_security_scheme(self, scheme: SecurityScheme, enforcer: SecurityEnforcer, *, default: bool = False):
        """
        Add a security scheme to be used by the API.

        :param scheme: a description of the security scheme
        :param enforcer: a function that checks the requirements of the security scheme
        :param default: should this be used as the default security scheme?
        """

        self._add_security_scheme(scheme, default)
        self.security_enforcers[scheme] = enforcer

    def _error_handler(self, exception: Exception):
        response = self.exception_to_response(exception)

        if response is None:
            raise ValueError(f"Unexpected exception type `{type(exception).__name__}`") from exception

        info, code = response

        return jsonify(info.to_primitive()), code

    def _get_default_error_handlers(self) -> Sequence[ErrorHandler]:
        return self._default_error_handlers

    def _is_bound(self) -> bool:
        return self._app is not None

    def _get_spec(self):
        """
        Serves the OpenAPI specification
        """
        self._extract_specs()
        return jsonify(self.to_openapi_dict())

    _get_spec.apistrap_ignore = True

    def _get_ui(self):
        """
        Serves Swagger UI
        """
        return render_template("apidocs.html", apistrap=self)

    _get_ui.apistrap_ignore = True

    def _get_redoc(self):
        """
        Serves ReDoc
        """
        return render_template("redoc.html", apistrap=self)

    _get_redoc.apistrap_ignore = True

    def _decorate_view_handlers(self):
        self._extract_operations()

    def _extract_operations(self):
        if self._operations is not None:
            return

        # Gather Flask view functions
        self._operations = []

        for rule in self._app.url_map.iter_rules():
            # Skip Flask's internal static endpoint
            if rule.endpoint == "static":
                continue

            handler = self._app.view_functions[rule.endpoint]

            for method in rule.methods:
                if self._is_route_ignored(method, handler):
                    continue

                op = FlaskOperationWrapper(self, handler, self._get_decorators(handler), str(rule), method)
                self._operations.append(op)
                self._app.view_functions[rule.endpoint] = op.get_decorated_function()

    def _extract_specs(self):
        """
        Extract specification data from the Flask app and save it to the underlying Apispec object
        """
        self._extract_operations()

        for op in self._operations:
            url = str(op.url_rule)
            for arg in re.findall("(<([^<>]*:)?([^<>]*)>)", url):
                url = url.replace(arg[0], "{%s}" % arg[2])

            self.spec.path(url, {op.method.lower(): op.get_openapi_spec()})

        self._specs_extracted = True

    def http_error_handler(self, exception: Exception):
        """
        A handler for Flask HTTPException objects
        :param exception: the exception raised due to a client error
        :return: a response object
        """

        if not isinstance(exception, HTTPException):
            raise ValueError()  # pragma: no cover

        logging.exception(exception)
        return ErrorResponse(dict(message=exception.description))

    def error_handler(self, exception):
        """
        The default handler for API client errors
        :param exception: the exception raised due to a client error
        :return: a response object
        """

        if self._app.debug:
            logging.exception(exception)
            return ErrorResponse(dict(message=str(exception), debug_data=format_exception(exception)))

        return ErrorResponse(dict(message=str(exception)))

    def internal_error_handler(self, exception):
        """
        The default handler for API server errors. It will hide the details when not in debug mode.
        :param exception: the exception raised due to a server error
        :return: a response object
        """

        logging.exception(exception)

        if self._app.debug:
            return ErrorResponse(dict(message=str(exception), debug_data=format_exception(exception)))

        return ErrorResponse(dict(message="Internal server error"))



from typing import Any, Mapping

from schematics import Model
from schematics.types import BaseType, DictType, StringType


class ErrorResponse(Model):
    """
    An error message wrapper
    """

    message: str = StringType(required=True)
    debug_data: Mapping[str, Any] = DictType(BaseType, required=False, serialize_when_none=False)


class EmptyResponse(Model):
    """
    An empty message wrapper
    """



from __future__ import annotations

from datetime import datetime
from io import BytesIO
from typing import TYPE_CHECKING, Dict, Union
from typing.io import BinaryIO

from schematics.exceptions import BaseError, CompoundError, ValidationError
from schematics.types import BaseType, FloatType, ListType, ModelType, PolyModelType

if TYPE_CHECKING:  # pragma: no cover
    try:
        from aiohttp import StreamReader
    except ImportError:

        class StreamReader:
            pass


class FileResponse:
    """
    File response used instead of `flask.send_file`.
    Note that MIME type should preferably be handled by `responds_with` decorator.
    """

    def __init__(
        self,
        filename_or_fp: Union[str, BinaryIO, BytesIO, StreamReader],
        as_attachment: bool = False,
        attachment_filename: str = None,
        add_etags: bool = True,
        cache_timeout: int = None,
        conditional: bool = False,
        last_modified: Union[datetime, int] = None,
        mimetype=None,
    ):
        self.filename_or_fp = filename_or_fp
        self.as_attachment = as_attachment
        self.attachment_filename = attachment_filename
        self.add_etags = add_etags
        self.cache_timeout = cache_timeout
        self.conditional = conditional
        self.last_modified = last_modified
        self.mimetype = mimetype


class TupleType(ListType):
    """A field for storing a fixed length tuple of items. The values must conform to the type
    specified by the ``field`` parameter.

    Use it like this::

        ...
        coords = TupleType(IntType, 3, required=False)

    """

    primitive_type = list
    native_type = tuple

    def __init__(self, field, length, **kwargs):
        self.field = self._init_field(field, kwargs)
        super().__init__(field, min_size=length, max_size=length, **kwargs)

    def _repr_info(self):
        return f"TupleType({self.field.__class__.__name__}, {self.max_size})"

    def _mock(self, context=None):
        return tuple([self.field._mock(context) for _ in range(self.max_size)])

    def _convert(self, value, context, safe=False):
        self._coerce(value)

        data = []
        errors = {}
        for i, val in enumerate(value):
            try:
                data.append(context.field_converter(self.field, val, context))
            except BaseError as exc:
                errors[i] = exc

        if errors:
            raise CompoundError(errors)
        return tuple(data)

    def _export(self, tup_instance, format, context):
        """The field export_level is ignored so it won't change tuple size / sequence of its elements."""
        data = []
        for value in tup_instance:
            shaped = self.field.export(value, format, context)
            data.append(shaped)
        return data


class NonNanFloatType(FloatType):
    """
    FloatType replacing NaN to zeros when transformed to JSON.
    Good for endpoint responses since JSON doesn't support NaNs but Python does.
    """

    _NOT_SET_VALUE = object()

    def __init__(self, default_value: Union[float, object] = _NOT_SET_VALUE, **kwargs):
        self._default_value = default_value
        super().__init__(**kwargs)

    def to_native(self, value, context=None):
        try:
            import numpy as np
        except ImportError as ex:  # pragma: no cover
            raise ImportError("NonNanFloatType requires Numpy.") from ex

        if np.isreal(value) and np.isscalar(value) and np.isnan(value):  # detect NaN
            if self._default_value == NonNanFloatType._NOT_SET_VALUE:  # user hasn't set the default value
                raise ValidationError("The value is NaN")
            value = self._default_value
        return super().to_native(value, context)


class DiscriminatedModelType(PolyModelType):
    """
    A Schematics field that converts to an "anyOf" model with a discriminator in OpenAPI schema.
    """

    def __init__(self, model_name: str, discriminator_field_name: str, type_map: Dict[str, ModelType], **kwargs):
        super().__init__([*type_map.values()], **kwargs)
        self._type_map = type_map
        self._discriminator_field_name = discriminator_field_name
        self._model_name = model_name

    @property
    def type_map(self):
        return self._type_map

    @property
    def model_name(self):
        return self._model_name

    @property
    def discriminator_field_name(self):
        return self._discriminator_field_name


class AnyType(BaseType):
    """
    A Schematics field that performs no validation/conversion and converts to an empty type in OpenAPI schema
    (which means "any").
    """



from setuptools import find_packages, setup


setup(name='apistrap',
      version='0.9.11',
      description='Iterait REST API utilities',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Operating System :: Unix',
          'Programming Language :: Python :: Implementation :: CPython',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.7',
      ],
      keywords='api rest http',
      url='https://github.com/iterait/apistrap',
      author=['Iterait a.s.', 'Cognexa Solutions s.r.o.'],
      author_email='hello@iterait.com',
      license='MIT',
      packages=find_packages('.', exclude='tests'),
      package_data={
          'apistrap': ['templates/*.html']
      },
      zip_safe=False,
      setup_requires=['pytest-runner'],
      tests_require=['pytest', 'pytest-mock', 'pytest-flask', 'pytest-aiohttp', 'flask<=1.1.4', 'aiohttp', 'numpy'],
      install_requires=['apispec==1.2', 'schematics', 'more_itertools', 'Werkzeug', 'jinja2', 'docstring_parser>=0.5'],
      extras_require={
          'flask': ['flask<=1.1.4'],
          'aiohttp': ['aiohttp'],
          'NonNanFloatType': ['numpy']
      })



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional, Sequence, Type, Union

from schematics import Model

from apistrap.tags import TagData

if TYPE_CHECKING:  # pragma: no cover
    from apistrap.extension import SecurityScheme
    from apistrap.utils import StringLike


class IgnoreDecorator:
    """
    Marks an endpoint as ignored so that Apistrap doesn't include it in the specification.
    """


@dataclass(frozen=True)
class IgnoreParamsDecorator:
    """
    Marks specified function parameters as ignored for the purposes of generating a specification
    """

    ignored_params: Sequence[str]


@dataclass(frozen=True)
class RespondsWithDecorator:
    """
    Specifies the format of the response. The response is automatically validated by Apistrap.
    """

    response_class: Type[Model]
    code: int = 200
    description: Optional[str] = None
    mimetype: Optional[str] = None


@dataclass(frozen=True)
class AcceptsDecorator:
    """
    Specifies the format of the request body and injects it as an argument to the view handler.
    The destination parameter must be annotated with a corresponding type.
    """

    request_class: Type[Model]
    mimetypes: Sequence[str] = ("application/json",)


@dataclass(frozen=True)
class AcceptsFileDecorator:
    """
    Declares that an endpoint accepts a file upload as the request body.
    """

    mime_type: str


@dataclass(frozen=True)
class AcceptsQueryStringDecorator:
    """
    Declares that an endpoint accepts query string parameters.
    """

    parameter_names: Sequence[str]


@dataclass(frozen=True)
class TagsDecorator:
    """
    Adds tags to the OpenAPI specification of the decorated view function.
    """

    tags: Sequence[Union[str, TagData]]


@dataclass(frozen=True)
class SecurityDecorator:
    """
    Enforces user authentication and authorization.
    """

    scopes: Sequence[StringLike]
    security_scheme: Optional[SecurityScheme] = None
